#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Mapycs (Matics Python API) usage example """

import os
import re

from mapycs import Dataframe
from mapycs import CatStruct
import test_eval

PATH = os.environ.get(r'CORPUS_DEV_P') + r"/audio"
NORM_PATH = PATH + r"/misc/normalization/normalize-repere.pl"
NGLM_PATH = PATH + r"/misc/nglm/fr.nglm"
REF_PATH = PATH + r"/ref/trs"
REF_PATTERN = r"^.*\.trs$"
HYP_PATH = PATH + r"/hyp/ctm"
HYP_PATTERN = r"^.*\.ctm$"
REF_PARAMS = {
    "language": "fre",
    "spk_local": "regexp",
    "spk_local_regexp": "^(Inconnu_|speaker#)"
}


def print_results(eval_df):

    """ Print the evaluation results """

    categories = [["global.system"], ["global.lang", "speaker.type"]]
    variables = ["wer ci.subst", "wer ci.insert", "wer ci.delete",
                 "wer ci.wer ci"]
    results = CatStruct(eval_df, categories, variables)
    print(results)


def get_refs():

    """ Populate reference files list """

    ref_files = []
    name_pattern = re.compile(REF_PATTERN)
    for file_name in os.listdir(REF_PATH):
        if name_pattern.match(file_name) is None:
            continue
        file_path = os.path.join(REF_PATH, file_name)
        ref_files.append(file_path)
    ref_files.sort()
    return ref_files


def get_hyps():

    """ Populate hypothesis files list """

    hyp_files = []
    name_pattern = re.compile(HYP_PATTERN)
    for file_name in os.listdir(HYP_PATH):
        if name_pattern.match(file_name) is None:
            continue
        file_path = os.path.join(HYP_PATH, file_name)
        hyp_files.append(file_path)
    hyp_files.sort()
    return hyp_files


def main():

    """ Dataframe usage examples """

    # Ref
    print("=" * 80)
    print("Loading ref")
    ref = Dataframe()
    ref_files = get_refs()
    for ref_file in ref_files:
        print("\t{}".format(os.path.basename(ref_file)))
        ref.load_annot(ref_file, REF_PARAMS)
    file_path = "/tmp/ref.dtf"
    print("-" * 80)
    print("Saving to {}".format(file_path))
    ref.save(file_path)
    print("-" * 80)
    print(ref)

#    # Hyp
    print("=" * 80)
    print("Loading hyp")
    hyp = Dataframe()
    hyp_files = get_hyps()
    for hyp_file in hyp_files:
        system = os.path.splitext(os.path.basename(hyp_file))[0]
        print("\t{}".format(system))
        params = {"systeme": system}
        params["language"] = "fre"
        hyp.load_annot(hyp_file, params)
    file_path = "/tmp/hyp.dtf"
    print("-" * 80)
    print("Saving to {}".format(file_path))
    hyp.save(file_path)
    print("-" * 80)
    print(hyp)

    cat = [["global.system"], ["global.lang", "speaker.type"]]
    var = ["wer ci.subst", "wer ci.insert", "wer ci.delete",
           "wer ci.wer ci"]
    params_list = [
        None,
        {
            "params": {"nglm": NGLM_PATH, "norm": NORM_PATH},
            "metrics": ["trans_ci", "trans_wer", "trans_nce",
                        "trans_ins", "trans_del", "trans_sub"]
        },
        {"metrics": ["trans_cer", "trans_nce"]},
        {"metrics": ["trans_cer", "trans_nce_"]},
        {"params": {"nglm": "aaaaaa", "bbbb": ""}}
    ]
    test_eval.test_eval("asr", params_list, ref, hyp, cat, var)


if __name__ == "__main__":

    main()
