#include "p3cs.h"
#include "p3df.h"

#include <Misc/strformat.h>

PyTypeObject *p3::catstruct::ptype;

PyMethodDef p3::catstruct::methods[] = {
    { "get_num_fact", get_num_fact, METH_NOARGS,
      "get_num_fact(self)\n--\n\nGet the number of factors"
    },
    { "get_fact_num_labels", get_fact_num_labels, METH_VARARGS,
      "get_fact_num_labels(self, fact, /)\n--\n\nGet the number of labels for a given factor"
    },
    { "get_fact_label", get_fact_label, METH_VARARGS,
      "get_fact_label(self, lbl, /)\n--\n\nGet a label for a a given factor"
    },
    { "get_num_var", get_num_var, METH_NOARGS,
      "get_num_var(self)\n--\n\nGet the number of variables"
    },
    { "get_var_label", get_var_label, METH_VARARGS,
      "get_var_label(self, var, /)\n--\n\nGet the label of a given variable"
    },
    { "get_var_labels", get_var_labels, METH_NOARGS,
      "get_var_labels(self)\n--\n\nGet the variables labels as a list"
    },
    { "get_num_vals", get_num_vals, METH_NOARGS,
      "get_num_vals(self)\n--\n\nGet the number of values"
    },
    { "get_value_int", get_value_int, METH_VARARGS,
      "get_value_int(self, var, val, /)\n--\n\nGet an integer value for a given variable"
    },
    { "get_value_dbl", get_value_dbl, METH_VARARGS,
      "get_value_dbl(self, var, val, /)\n--\n\nGet a float value for a given variable"
    },
    { "get_value_str", get_value_str, METH_VARARGS,
      "get_value_str(self, var, val, /)\n--\n\nGet a string value for a given variable"
    },
    { "get_value", get_value, METH_VARARGS,
      "get_value(self, var, val, /)\n--\n\nGet a value for a given variable"
    },
    { "get_cat", get_cat, METH_VARARGS,
      "get_cat(self, fact, val, /)\n--\n\nGet the factor category for a given value"
    },
    { "get_cat_num_labels", get_cat_num_labels, METH_VARARGS,
      "get_cat_num_labels(self, cat, /)\n--\n\nGet the number of labels for a given category"
    },
    { "get_cat_label", get_cat_label, METH_VARARGS,
      "get_cat_label(self, fact, cat, lbl, /)\n--\n\nGet a label for a given factor and category"
    },
    { "get_facts", get_facts, METH_NOARGS,
      "get_facts(self)\n--\n\nGet the factors labels as a list"
    },
    { "get_categories", get_categories, METH_VARARGS,
      "get_categories(self, val, /)\n--\n\nGet the categories labels as a list"
    },
    {}
};

int p3::catstruct::init(PyObject *self, PyObject *args, PyObject *)
{
    PyObject *odf, *categories, *values;
    if(!PyArg_ParseTuple(args, "O!OO", dataframe::ptype, &odf, &categories, &values))
        return 1;

    if(!PySequence_Check(categories) || !PySequence_Check(values)) {
        PyErr_BadArgument();
        return 1;
    }

    auto df = ((dataframe *)odf)->df;
    std::vector<std::vector<int>> cats;
    cats.resize(PySequence_Size(categories));
    for(Py_ssize_t i = 0; i != PySequence_Size(categories); i++) {
        auto subcat = PySequence_GetItem(categories, i);
        if(!PySequence_Check(subcat)) {
            PyErr_BadArgument();
            return 1;
        }
        cats[i].resize(PySequence_Size(subcat));
        for(Py_ssize_t j = 0; j != PySequence_Size(subcat); j++) {
            auto name = PySequence_GetItem(subcat, j);
            if(!PyUnicode_Check(name)) {
                PyErr_BadArgument();
                return 1;
            }
            auto n1 = tostr(name);
            int col = df->getIndiceColonneByKey(n1);
            if(col == -1) {
                PyErr_SetString(PyExc_RuntimeError, util::string_format("Nom de colonne inconnu '%s'", n1).c_str());
                return 1;
            }
            cats[i][j] = col;
            Py_DecRef(name);
        }
        Py_DecRef(subcat);
    }

    std::vector<int> vals;
    vals.resize(PySequence_Size(values));
    for(Py_ssize_t j = 0; j != PySequence_Size(values); j++) {
        auto name = PySequence_GetItem(values, j);
        if(!PyUnicode_Check(name)) {
            PyErr_BadArgument();
            return 1;
        }
        auto n1 = tostr(name);
        int col = df->getIndiceColonneByKey(n1);
        if(col == -1) {
            PyErr_SetString(PyExc_RuntimeError, util::string_format("Nom de colonne inconnu '%s'", n1).c_str());
            return 1;
        }
        vals[j] = col;
        Py_DecRef(name);
    }

    auto cdf = df->creationMatriceCorrespondance();
    cdf->load();
    ((catstruct *)self)->sr = cdf->creationStructuresCategoriesEtValeurs(cats, vals, COLLAPSE);
    delete cdf;
    return 0;
}

PyObject *p3::catstruct::str(PyObject *self)
{
    const auto &sr = ((catstruct *)self)->sr;
    std::string desc;
    for(const auto &l : sr.labelsSDC) {
        desc += '\t';
        for(size_t j=0; j != l.size(); j++) {
            if(j)
                desc += '|';
            desc += l[j];
        }
    }
    for(const auto &v : sr.labelsValeurs)
        desc += '\t' + v;

    auto nv = sr.vecteurValeurs.front().size();
    for(size_t v = 0; v != nv; v++) {
        desc += util::string_format("\n%d", v);

        for(size_t i=0; i != sr.vecteurSDC.size(); i++) {
            desc += '\t';
            const auto &f = sr.vecteurSDC[i];
            int cat = f.idCombos[v];
            for(size_t j=0; j != sr.labelsSDC[i].size(); j++) {
                if(j)
                    desc += ", ";
                auto        id = f.idCats[j][cat];
                if (-1 == id) desc += "(empty)";
                else desc += f.catLbls[j][id];
            }
        }
        for(const auto &vals : sr.vecteurValeurs) {
            const auto &val = vals[v];
            if(val.isString())
                desc += util::string_format("\t%s", val.toString());
            else if(val.isDouble())
                desc += util::string_format("\t%g", val.toDouble());
            else
                desc += "\t(empty)";
        }
    }
    return PyUnicode_FromString(desc.c_str());
}

bool p3::catstruct::reg(PyObject *module)
{
    return type_register<catstruct>(module, ptype, "CatStruct", "A data extraction structure and result");
}

PyObject *p3::catstruct::get_num_fact(PyObject *self, PyObject *)
{
    const auto &sr = ((catstruct *)self)->sr;
    return PyLong_FromSize_t(sr.labelsSDC.size());
}

PyObject *p3::catstruct::get_fact_num_labels(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int fact;
    if(!PyArg_ParseTuple(args, "i", &fact))
        return nullptr;
    return PyLong_FromSize_t(sr.labelsSDC[fact].size());
}

PyObject *p3::catstruct::get_fact_label(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int fact, lbl;
    if(!PyArg_ParseTuple(args, "ii", &fact, &lbl))
        return nullptr;
    if(fact >= int(sr.labelsSDC.size())) {
        std::string err = util::string_format("Incorrect factor number %d (max %d)", fact, sr.labelsSDC.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(lbl >= int(sr.labelsSDC[fact].size())) {
        std::string err = util::string_format("Incorrect label number %d (max %d)", lbl, sr.labelsSDC[fact].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyUnicode_FromString(sr.labelsSDC[fact][lbl].c_str());
}

PyObject *p3::catstruct::get_num_var(PyObject *self, PyObject *)
{
    const auto &sr = ((catstruct *)self)->sr;
    return PyLong_FromSize_t(sr.vecteurValeurs.size());
}

PyObject *p3::catstruct::get_var_label(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int var;
    if(!PyArg_ParseTuple(args, "i", &var))
        return nullptr;
    if(var >= int(sr.labelsValeurs.size())) {
        std::string err = util::string_format("Incorrect variable number %d (max %d)", var, sr.labelsValeurs.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyUnicode_FromString(sr.labelsValeurs[var].c_str());
}

PyObject *p3::catstruct::get_var_labels(PyObject *self, PyObject *)
{
    const auto &sr = ((catstruct *)self)->sr;
    auto res = PyList_New(sr.labelsValeurs.size());
    for(size_t i=0; i != sr.labelsValeurs.size(); i++)
        PyList_SetItem(res, i, PyUnicode_FromString(sr.labelsValeurs[i].c_str()));
    return res;
}

PyObject *p3::catstruct::get_num_vals(PyObject *self, PyObject *)
{
    const auto &sr = ((catstruct *)self)->sr;
    if (sr.vecteurValeurs.size() == 0)
    {
        PyErr_SetNone(PyExc_IndexError);
        return NULL;
    }
    return PyLong_FromSize_t(sr.vecteurValeurs.front().size());
}

PyObject *p3::catstruct::get_value_int(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int var, val;
    if(!PyArg_ParseTuple(args, "ii", &var, &val))
        return nullptr;
    if(var >= int(sr.vecteurValeurs.size())) {
        std::string err = util::string_format("Incorrect variable number %d (max %d)", var, sr.vecteurValeurs.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(val >= int(sr.vecteurValeurs[var].size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurValeurs[var].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyLong_FromLongLong(static_cast<long long>(sr.vecteurValeurs[var][val].toDouble()));
}

PyObject *p3::catstruct::get_value_dbl(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int var, val;
    if(!PyArg_ParseTuple(args, "ii", &var, &val))
        return nullptr;
    if(var >= int(sr.vecteurValeurs.size())) {
        std::string err = util::string_format("Incorrect variable number %d (max %d)", var, sr.vecteurValeurs.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(val >= int(sr.vecteurValeurs[var].size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurValeurs[var].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyFloat_FromDouble(sr.vecteurValeurs[var][val].toDouble());
}

PyObject *p3::catstruct::get_value_str(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int var, val;
    if(!PyArg_ParseTuple(args, "ii", &var, &val))
        return nullptr;
    if(var >= int(sr.vecteurValeurs.size())) {
        std::string err = util::string_format("Incorrect variable number %d (max %d)", var, sr.vecteurValeurs.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(val >= int(sr.vecteurValeurs[var].size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurValeurs[var].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyUnicode_FromString(sr.vecteurValeurs[var][val].toString().c_str());
}

PyObject *p3::catstruct::get_value(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int var, val;
    if(!PyArg_ParseTuple(args, "ii", &var, &val))
        return nullptr;
    if(var >= int(sr.vecteurValeurs.size())) {
        std::string err = util::string_format("Incorrect variable number %d (max %d)", var, sr.vecteurValeurs.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(val >= int(sr.vecteurValeurs[var].size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurValeurs[var].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    const auto &vv(sr.vecteurValeurs[var][val]);
    if (vv.isNan()) return PyUnicode_FromString("NAN");
    else if (vv.isNull()) return PyUnicode_FromString("NULL");
    else if (vv.isNegativeInf()) return PyUnicode_FromString("-INF");
    else if (vv.isPositiveInf()) return PyUnicode_FromString("+INF");
    else if (vv.isDouble()) return PyFloat_FromDouble(vv.toDouble());
    else if (vv.isString()) return PyUnicode_FromString(vv.toString().c_str());
    else return nullptr;
}

PyObject *p3::catstruct::get_cat(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int fact, val;
    if(!PyArg_ParseTuple(args, "ii", &fact, &val))
        return nullptr;
    if(fact >= int(sr.vecteurSDC.size())) {
        std::string err = util::string_format("Incorrect factor number %d (max %d)", fact, sr.vecteurSDC.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(val >= int(sr.vecteurSDC[fact].idCombos.size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurSDC[fact].idCombos.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyLong_FromLong(sr.vecteurSDC[fact].idCombos[val]);
}

PyObject *p3::catstruct::get_cat_num_labels(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int cat;
    if(!PyArg_ParseTuple(args, "i", &cat))
        return nullptr;
    if(cat >= int(sr.vecteurSDC.size())) {
        std::string err = util::string_format("Incorrect category number %d (max %d)", cat, sr.vecteurSDC.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyLong_FromSize_t(sr.vecteurSDC[cat].catLbls.size());
}

PyObject *p3::catstruct::get_cat_label(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int fact, cat, lbl;
    if(!PyArg_ParseTuple(args, "iii", &fact, &cat, &lbl))
        return nullptr;
    if(fact >= int(sr.vecteurSDC.size())) {
        std::string err = util::string_format("Incorrect factor number %d (max %d)", fact, sr.vecteurSDC.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    const auto &f = sr.vecteurSDC[fact];
    if(lbl >= int(f.idCats.size())) {
        std::string err = util::string_format("Incorrect label number %d (max %d)", lbl, f.idCats.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    if(cat >= int(f.idCats[lbl].size())) {
        std::string err = util::string_format("Incorrect category number %d (max %d)", cat, f.idCats[lbl].size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    return PyUnicode_FromString(f.catLbls[lbl][f.idCats[lbl][cat]].c_str());
}

PyObject *p3::catstruct::get_facts(PyObject *self, PyObject *)
{
    const auto &sr = ((catstruct *)self)->sr;
    auto res = PyList_New(sr.labelsSDC.size());
    for(size_t i=0; i != sr.labelsSDC.size(); i++) {
        auto entry = PyList_New(sr.labelsSDC[i].size());
        for(size_t j=0; j != sr.labelsSDC[i].size(); j++)
            PyList_SetItem(entry, j, PyUnicode_FromString(sr.labelsSDC[i][j].c_str()));
        PyList_SetItem(res, i, entry);
    }
    return res;
}

PyObject *p3::catstruct::get_categories(PyObject *self, PyObject *args)
{
    const auto &sr = ((catstruct *)self)->sr;
    int val;
    if(!PyArg_ParseTuple(args, "i", &val))
        return nullptr;
    if(val >= int(sr.vecteurSDC[0].idCombos.size())) {
        std::string err = util::string_format("Incorrect value number %d (max %d)", val, sr.vecteurSDC[0].idCombos.size()-1);
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    auto res = PyList_New(sr.vecteurSDC.size());
    for(size_t i=0; i != sr.vecteurSDC.size(); i++) {
        const auto &f = sr.vecteurSDC[i];
        int cat = f.idCombos[val];
        auto entry = PyList_New(f.catLbls.size());
        for(size_t j=0; j != sr.labelsSDC[i].size(); j++)
            PyList_SetItem(entry, j, PyUnicode_FromString(f.catLbls[j][f.idCats[j][cat]].c_str()));
        PyList_SetItem(res, i, entry);
    }
    return res;
}
