#include "p3df.h"

#include <Dataframe/csv.h>
#include <Format/formats.h>

#include <list>

PyTypeObject *p3::dataframe::ptype;

PyMethodDef p3::dataframe::methods[] = {
    { "get_formats", get_formats, METH_NOARGS|METH_STATIC,
      "get_formats()\n--\n\nReturn the supported annotation formats as a dictionary."
    },
    { "from_file", from_file, METH_VARARGS|METH_STATIC,
      "from_file(path, /)\n--\n\nLoad a Dataframe from a file path.\n\nparam path: dtf file path.\n"
    },
    { "load_annot", load_annot, METH_VARARGS,
      "load_annot(self, path, params=None, /)\n--\n\nLoad data from an annotation file into the dataframe.\n"
      "\nNew data will appended to the current content.\n\n"
      ":param path: source data file path\n"
      ":param params: annotation parameters\n"
    },
    { "load_annot_with_format", load_annot_with_format, METH_VARARGS,
      "load_annot_with_format(self, path, category, type, params=None, /)\n--\n\nLoad data from an annotation file of a specific format into the dataframe.\n"
      "\nNew data will appended to the current content.\n\n"
      ":param path: source data file path\n"
      ":param category: format category\n"
      ":param type: format type\n"
      ":param params: annotation parameters\n"
    },
    { "save", save, METH_VARARGS,
      "save(path, /)\n--\n\nSave the dataframe contents to a file.\n\nparam path: dtf file path.\n"
    },
    { "normalize", normalize, METH_VARARGS,
      "normalize(self, old_key, new_key, new_reduce, script_path, /)\n--\n\nAdd a new column for normalized content..\n"
      "\nNew data will appended to the current content.\n\n"
      ":param old_key: source text column key\n"
      ":param new_key: new column key\n"
      ":param new_reduce: new column reduce\n"
      ":param script_path: normalization script file path\n"
    },
    { "get_lines_count", get_lines_count, METH_NOARGS,
      "get_lines_count(self)\n--\n\n:returns: lines count.\n"
    },
    { "get_cols_count", get_cols_count, METH_NOARGS,
      "get_cols_count(self)\n--\n\n:returns: columns count.\n"
    },
    { "get_cols_keys", get_cols_keys, METH_NOARGS,
      "get_cols_keys(self)\n--\n\n:returns: list of columns names.\n"
    },
    { "get_val_val_num", get_val_val_num, METH_VARARGS,
      "get_val_val_num(self, line, col, /)\n--\n\nReturns the contents of a numeric value column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    { "get_val_val_str", get_val_val_str, METH_VARARGS,
      "get_val_val_str(self, line, col, /)\n--\n\nReturns the contents of a string value column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    { "get_val_lab_num", get_val_lab_num, METH_VARARGS,
      "get_val_lab_num(self, line, col, /)\n--\n\nReturns the contents of a numeric label column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    { "get_val_lab_str", get_val_lab_str, METH_VARARGS,
      "get_val_lab_str(self, line, col, /)\n--\n\nReturns the contents of a string label column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    { "get_val_line", get_val_line, METH_VARARGS,
      "get_val_line(self, line, col, /)\n--\n\nReturns the reference line for a column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the reference line number\n"
    },
    { "get_val_expr", get_val_expr, METH_VARARGS,
      "get_val_expr(self, line, col, /)\n--\n\nReturns the value for a reference column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the reference line number\n"
    },
    { "get_val", get_val, METH_VARARGS,
      "get_val(self, line, col, /)\n--\n\nReturns the value for a column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    { "add_medias", add_medias, METH_VARARGS,
      "add_medias(self, media_paths, /)\n--\n\nReturns the value for a column at a given position.\n\n:param line: line number\n:param col:column number or key\n:returns: the value\n"
    },
    {}
};

int p3::dataframe::init(PyObject *self, PyObject*, PyObject*)
{
    ((dataframe *)self)->set(new Dataframe(false));
    return 0;
}

PyObject *p3::dataframe::str(PyObject *self)
{
    auto df = ((dataframe *)self)->df;
    std::string desc;
    desc = util::string_format("%d lines, %d columns", df->lineCount(), df->columnCount());
    for(int i=0; i != df->columnCount(); i++)
        desc += util::string_format("\n%d\t%s", i, df->getColumnKey(i));
    return PyUnicode_FromString(desc.c_str());
}

bool p3::dataframe::reg(PyObject *module)
{
    return type_register<dataframe>(module, ptype, "Dataframe", "A Matics Dataframe");
}

std::unique_ptr<bson::document> p3::dataframe::parse_params(PyObject *params)
{
    auto par = bson::make_document();
    if(!params || params == Py_None)
        return par;

    if(!PyMapping_Check(params)) {
        PyErr_BadArgument();
        return nullptr;
    }

    auto items = PyMapping_Items(params);
    Py_ssize_t icount = PySequence_Length(items);
    for(Py_ssize_t i=0; i != icount; i++) {
        auto item = PySequence_GetItem(items, i);
        auto key = PyTuple_GetItem(item, 0);
        if(!PyUnicode_Check(key)) {
            PyErr_BadArgument();
            return nullptr;
        }
        auto value = PyTuple_GetItem(item, 1);
        if(!PyUnicode_Check(value)) {
            PyErr_BadArgument();
            return nullptr;
        }
        par->insert(p3::tostr(key), p3::tostr(value));
        Py_DECREF(item);
    }
    Py_DECREF(items);
    return par;
}

PyObject *p3::dataframe::get_formats(PyObject *, PyObject *)
{
    auto res = PyDict_New();
    for(unsigned i=0; !all_formats[i].is_end(); i++) {
        const annotation::Builder &format = all_formats[i];
        auto entry = PyDict_New();
        auto cat = PyUnicode_FromString(format.category());
        PyDict_SetItemString(entry, "category", cat);
        Py_DECREF(cat);
        auto params = PyList_New(0);
        for(const char **p = format.used_params(); *p; p++) {
            auto name = PyUnicode_FromString(*p);
            PyList_Append(params, name);
            Py_DECREF(name);
        }
        PyDict_SetItemString(entry, "params", params);
        Py_DECREF(params);

        PyDict_SetItemString(res, format.type(), entry);
        Py_DECREF(entry);
    }
    return res;
}

PyObject *p3::dataframe::from_file(PyObject *, PyObject *args)
{
    const char *path;
    if(!PyArg_ParseTuple(args, "s", &path))
        return nullptr;

    const void *bdata;
    std::size_t bsize;
    try {
        filemgr::map_ro(path, bdata, bsize);
    } catch(const filemgr::filesystem_error &e) {
        PyErr_SetString(PyExc_RuntimeError, e.errstr.c_str());
        return nullptr;
    }

    std::string err;
    Dataframe *df = csv_read(false, bdata, bsize, err);
    if(!df) {
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }

    auto ndf = (dataframe *)(PyObject_CallFunction((PyObject *)ptype, ""));
    delete ndf->df;
    ndf->df = df;
    return &ndf->ob_base;
}

PyObject *p3::dataframe::load_annot(PyObject *self, PyObject *args)
{
    auto df = (dataframe *)self;
    const char *path;
    PyObject *params = nullptr;
    if(!PyArg_ParseTuple(args, "s|O", &path, &params))
        return nullptr;

    auto par = parse_params(params);
    if(!par)
        return nullptr;

    std::shared_ptr<filemgr::fmap> map;
    try {
        map = std::make_shared<filemgr::fmap>(path);
    } catch(const filemgr::filesystem_error &e) {
        PyErr_SetString(PyExc_RuntimeError, e.errstr.c_str());
        return nullptr;
    }

    annotation::errors_t errors(map.get());
    auto data = annotation::load(path, map, errors);
    if(!data) {
        std::string err;
        for(const auto &errtype : errors.errors)
            if(!errtype.errors.empty()) {
                err += util::string_format("Erreur au chargement en tant que %1$s:\n", errtype.type);
                for(const auto &errentry : errtype.errors)
                    err += util::string_format("%1$d:%2$d: %3$s\n", errentry.line, errentry.column, errentry.message);        }
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    data->append_dataframe(*df->df, annotation::param_t(par.get()));
    Py_RETURN_NONE;
}

PyObject *p3::dataframe::load_annot_with_format(PyObject *self, PyObject *args)
{
    auto df = (dataframe *)self;
    const char *path;
    const char *cat;
    const char *type;
    PyObject *params = nullptr;
    if(!PyArg_ParseTuple(args, "sss|O", &path, &cat, &type, &params))
        return nullptr;

    auto par = parse_params(params);
    if(!par)
        return nullptr;

    std::shared_ptr<filemgr::fmap> map;
    try {
        map = std::make_shared<filemgr::fmap>(path);
    } catch(const filemgr::filesystem_error &e) {
        PyErr_SetString(PyExc_RuntimeError, e.errstr.c_str());
        return nullptr;
    }

    annotation::errors_t errors(map.get());
    auto data = annotation::load(path, map, cat, type, errors);
    if(!data) {
        std::string err;
        for(const auto &errtype : errors.errors)
            if(!errtype.errors.empty()) {
                err += util::string_format("Erreur au chargement en tant que %1$s:\n", errtype.type);
                for(const auto &errentry : errtype.errors)
                    err += util::string_format("%1$d:%2$d: %3$s\n", errentry.line, errentry.column, errentry.message);        }
        PyErr_SetString(PyExc_RuntimeError, err.c_str());
        return nullptr;
    }
    data->append_dataframe(*df->df, annotation::param_t(par.get()));
    Py_RETURN_NONE;
}

PyObject *p3::dataframe::save(PyObject *self, PyObject *args)
{
    auto df = (dataframe *)self;
    const char *path;
    if(!PyArg_ParseTuple(args, "s", &path))
        return nullptr;

    df->df->sauvegarderDataframe(path);

    Py_RETURN_NONE;
}

PyObject *p3::dataframe::normalize(PyObject *self, PyObject *args)
{
    auto df = (dataframe *)self;
    const char *old_key, *new_key, *new_reduce, *path;
    if(!PyArg_ParseTuple(args, "ssss", &old_key, &new_key, &new_reduce, &path))
        return nullptr;
    int oid = df->df->getIndiceColonneByKey(old_key);
    if(oid == -1) {
        PyErr_SetString(PyExc_RuntimeError, util::string_format("Colonne %s non trouvée", old_key).c_str());
        return nullptr;
    }
    int nid = df->df->getIndiceColonneByKey(new_key);
    if(nid != -1) {
        PyErr_SetString(PyExc_RuntimeError, util::string_format("La colonne %s existe déjà", new_key).c_str());
        return nullptr;
    }

    std::string ncol = new_key;
    auto pos = ncol.find('.');
    if(pos == std::string::npos) {
        PyErr_SetString(PyExc_RuntimeError, "Le nom de colonne doit avoir la forme groupe.nom");
        return nullptr;
    }
    std::string groupe = ncol.substr(0, pos);
    std::string nom = ncol.substr(pos+1);
    bool err, already_present = false;
    std::string errors;
    df->df->appliquerScriptNormalisation(nom, groupe, oid, path, new_reduce, err, errors, already_present);
    if(err) {
        PyErr_SetString(PyExc_RuntimeError, errors.c_str());
        return nullptr;
    }
    Py_RETURN_NONE;
}

PyObject *p3::dataframe::add_medias(PyObject *self, PyObject *args)
{
    // Arguments
    PyObject        *media_list;
    if (! PyArg_ParseTuple(args, "O", &media_list))
        Py_RETURN_NONE;
    if (! PyList_Check(media_list))
    {
        PyErr_SetString(PyExc_TypeError, util::string_format("Expected list type").c_str());
        Py_RETURN_NONE;
    }
    std::list<std::string>      media_paths;
    for (Py_ssize_t mn = 0; mn < PyList_Size(media_list); mn++)
    {
        auto        i = PyList_GetItem(media_list, mn);
        media_paths.push_back(PyUnicode_AsUTF8(i));
    }
    auto        df = ((dataframe *)self)->df;
    std::unordered_map<std::string, std::string>        medias;
    annotation::media_extract(df, medias);
    for (const auto &media_path : media_paths)
    {
        Corpus::FileInfo        finfo;
        try
        {
            finfo.map = std::make_shared<filemgr::fmap>(media_path);
        }
        catch(filemgr::filesystem_error &fe)
        {
            fprintf(stderr, "Error on accessing %s: %s\n", fe.path.c_str(), fe.errstr.c_str());
            exit(1);
        }
        finfo.info->insert("size", double(finfo.map->size));
        if (! finfo.map->size ||
            ! ffmpeg::probe(media_path, finfo.map, finfo.info.get())) continue;
        medias[annotation::path_to_simplified_file_name(media_path)] = media_path;
    }
    annotation::media_integrate(df, medias);
    Py_RETURN_NONE;
}

PyObject *p3::dataframe::get_lines_count(PyObject *self, PyObject *)
{
    auto df = ((dataframe *)self)->df;
    return PyLong_FromLong(df->lineCount());
}

PyObject *p3::dataframe::get_cols_count(PyObject *self, PyObject *)
{
    auto df = ((dataframe *)self)->df;
    return PyLong_FromLong(df->columnCount());
}

PyObject *p3::dataframe::get_cols_keys(PyObject *self, PyObject *)
{
    auto df = ((dataframe *)self)->df;
    auto res = PyList_New(df->columnCount());
    for(int i=0; i != df->columnCount(); i++)
        PyList_SetItem(res, i, PyUnicode_FromString(df->getColumnKey(i).c_str()));
    return res;
}

bool p3::dataframe::get_linecol(PyObject *args, Dataframe *df, int &line, int &col)
{
    const char *scol;
    if(PyArg_ParseTuple(args, "ii", &line, &col))
        return true;
    PyErr_Clear();
    if(!PyArg_ParseTuple(args, "is", &line, &scol))
        return false;
    col = df->getIndiceColonneByKey(scol);
    if(col == -1) {
        PyErr_SetString(PyExc_RuntimeError, "Nom de colonne inconnu");
        return false;
    }
    return true;
}

PyObject *p3::dataframe::get_val_val_num(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if((df->getColumnType(col) & (IsLabel|IsString)) != 0) {
        PyErr_SetString(PyExc_RuntimeError, "Mauvais type de colonne");
        return nullptr;
    }
    if(df->isEmptyValue(line, col))
        Py_RETURN_NONE;
    return PyFloat_FromDouble(df->getValueValNum(line, col));
}

PyObject *p3::dataframe::get_val_val_str(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if((df->getColumnType(col) & (IsLabel|IsString)) != IsString) {
        PyErr_SetString(PyExc_RuntimeError, "Mauvais type de colonne");
        return nullptr;
    }
    if(df->isEmptyValue(line, col))
        Py_RETURN_NONE;
    return PyUnicode_FromString(df->getValueValStr(line, col).c_str());
}

PyObject *p3::dataframe::get_val_lab_num(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if((df->getColumnType(col) & (IsLabel|IsString)) != IsLabel) {
        PyErr_SetString(PyExc_RuntimeError, "Mauvais type de colonne");
        return nullptr;
    }
    if(df->isEmptyValue(line, col))
        Py_RETURN_NONE;
    return PyFloat_FromDouble(df->getValueLabNum(line, col));
}

PyObject *p3::dataframe::get_val_lab_str(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if((df->getColumnType(col) & (IsLabel|IsString)) != (IsLabel|IsString)) {
        PyErr_SetString(PyExc_RuntimeError, "Mauvais type de colonne");
        return nullptr;
    }
    if(df->isEmptyValue(line, col))
        Py_RETURN_NONE;
    return PyUnicode_FromString(df->getValueLabStr(line, col).c_str());
}

PyObject *p3::dataframe::get_val_line(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    return PyLong_FromLong(df->getValueLine(line, col));
}

PyObject *p3::dataframe::get_val_expr(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if((df->getColumnType(col) & IsExpression) != IsExpression) {
        PyErr_SetString(PyExc_RuntimeError, "Mauvais type de colonne");
        return nullptr;
    }

    GroupeColonnes *group = new GroupeColonnes();
    group->m_numLignes.push_back(line);
    group->m_nbElements = 1;
    auto val = CorrespondanceDataframe(df).recupererValeurExpression(group, col).m_valeur;
    delete group;

    if((df->getColumnType(col) & IsString) == 0)
        return PyFloat_FromDouble(val.toDouble());
    else
        return PyUnicode_FromString(val.toString().c_str());
}

PyObject *p3::dataframe::get_val(PyObject *self, PyObject *args)
{
    auto df = ((dataframe *)self)->df;
    int line, col;
    if(!get_linecol(args, df, line, col))
        return nullptr;
    if(df->getColumnType(col) & IsExpression) {
        GroupeColonnes *group = new GroupeColonnes();
        group->m_numLignes.push_back(line);
        group->m_nbElements = 1;
        auto val = CorrespondanceDataframe(df).recupererValeurExpression(group, col).m_valeur;
        delete group;

        if((df->getColumnType(col) & IsString) == 0)
            return PyFloat_FromDouble(val.toDouble());
        else
            return PyUnicode_FromString(val.toString().c_str());
    }

    if(df->isEmptyValue(line, col))
        Py_RETURN_NONE;

    switch(df->getColumnType(col)) {
    case 0: return PyFloat_FromDouble(df->getValueValNum(line, col));
    case IsString: return PyUnicode_FromString(df->getValueValStr(line, col).c_str());
    case IsLabel: return PyFloat_FromDouble(df->getValueLabNum(line, col));
    case IsLabel|IsString: return PyUnicode_FromString(df->getValueLabStr(line, col).c_str());
    }
    Py_RETURN_NONE;
}
