#include "p3eval.h"
#include "p3df.h"

#include <Evaluations/eval.h>
#include <Misc/strformat.h>

#include <sstream>

PyTypeObject *p3::eval::ptype;

PyMethodDef p3::eval::methods[] = {
    { "evals", evals, METH_NOARGS|METH_STATIC,
      "evals()\n--\n\nReturn the supported evaluations as a dictionary."
    },
    { "eval_params", eval_params, METH_VARARGS|METH_STATIC,
      "eval_params()\n--\n\n"
    },
    { "get_dtfs", get_dtfs, METH_NOARGS,
      "get_dtfs(self)\n--\n\n:returns: Evaluation results dataframe"
    },
    { "get_res_df", get_res_df, METH_NOARGS,
      "get_res_df(self)\n--\n\n:returns: Evaluation results dataframe"
    },
    { "get_err", get_err, METH_NOARGS, ""},
    {}
};
bool p3::eval::reg(PyObject *module)
{
    return type_register<eval>(module, ptype, "Eval", "A Matics evaluation");
}

int p3::eval::init(PyObject *self, PyObject *args, PyObject *)
{
    // Arguments
    const char*     type_id;
    PyObject        *p;
    PyObject        *rdf, *hdf, *sdf(nullptr);
    if (! PyArg_ParseTuple(args, "sOO!O!|O!", &type_id, &p,
                           dataframe::ptype, &rdf, dataframe::ptype, &hdf, dataframe::ptype, &sdf))
        return -1;

    // Eval
    auto        obj = (eval *) self;
    obj->eva = ::Eval::create(::Eval::type_from_id(type_id));

    // Eval parameters
    if (p != Py_None)
    {
        if (! PyDict_Check(p))
        {
            PyErr_SetString(PyExc_RuntimeError, util::string_format("Expected dictionnary type").c_str());
            return -1;
        }
        // Metrics
        auto        key = PyUnicode_FromString("metrics");
        if (-1 == PyDict_Contains(p, key))
        {
            PyErr_SetString(PyExc_RuntimeError, util::string_format("Metrics ?").c_str());
            return -1;
        }
        std::set<std::string>     metrics;
        if (1 == PyDict_Contains(p, key))
        {
            auto        lst = PyDict_GetItemWithError(p, key);
            Py_DECREF(key);

            for (Py_ssize_t mn = 0; mn < PyList_Size(lst); mn++)
            {
                auto        i = PyList_GetItem(lst, mn);
                metrics.insert(PyUnicode_AsUTF8(i));
            }
        }
        // Params
        auto        key2 = PyUnicode_FromString("params");
        if (-1 == PyDict_Contains(p, key2))
        {
            PyErr_SetString(PyExc_RuntimeError, util::string_format("Params ?").c_str());
            return -1;
        }
        std::map<std::string, ::Eval::ParamValue> params;
        if (1 == PyDict_Contains(p, key2))
        {
            auto        dict = PyDict_GetItemWithError(p, key2);
            Py_DECREF(key2);
            if (! PyDict_Check(dict))
            {
                PyErr_SetString(PyExc_RuntimeError, util::string_format("Expected dictionnary type (2)").c_str());
                return -1;
            }
            PyObject *pkey, *value;
            Py_ssize_t pos = 0;
            while (PyDict_Next(dict, &pos, &pkey, &value))
            {
                Eval::ParamType     t;
                if (! obj->eva->param_type(PyUnicode_AsUTF8(pkey), t))
                {
                    std::string     err = std::string("Unknown param '") + PyUnicode_AsUTF8(pkey) + "'";
                    PyErr_SetString(PyExc_RuntimeError, util::string_format(err).c_str());
                    return -1;
                }
                switch (t)
                {
                case Eval::FLOAT:
                {
                    auto i = PyFloat_AsDouble(value);
                    if (i == -1 && PyErr_Occurred()) {
                        return -1;
                    }
                    params.emplace(PyUnicode_AsUTF8(pkey), i);
                    break;
                }
                case Eval::INT:
                {
                    auto i = PyLong_AsLong(value);
                    if (i == -1 && PyErr_Occurred()) {
                        return -1;
                    }
                    params.emplace(PyUnicode_AsUTF8(pkey), (int) i);
                    break;
                }
                case Eval::PATH:
                {
                    auto i = PyUnicode_AsUTF8(value);
                    if (i == NULL && PyErr_Occurred()) {
                        return -1;
                    }
                    params.emplace(PyUnicode_AsUTF8(pkey), i);
                    break;
                }
                }
            }
        }
        if (nullptr == obj->eva)
        {
            PyErr_SetString(PyExc_RuntimeError, util::string_format("Unknown eval type '%s'", type_id).c_str());
            return -1;
        }
        auto        merr = obj->eva->set_params(params, metrics);
        if (! merr.ok)
        {
            std::string     err_str = merr.title + ": " + merr.desc;
            PyErr_SetString(PyExc_RuntimeError,
                            util::string_format(err_str).c_str());
            return -1;
        }
    }
    // Dataframes
    auto r_df = ((dataframe *) rdf)->df;
    auto h_df = ((dataframe *) hdf)->df;
    Dataframe       *s_df = nullptr;
    if (nullptr != sdf) s_df = ((dataframe *) sdf)->df;
    obj->eva->set_dtf(r_df, h_df, s_df);

    auto        results = obj->eva->result(obj->err);
    if (results.size() == 0)
    {
        PyErr_SetString(PyExc_RuntimeError,
                        util::string_format(obj->err).c_str());
        return -1;
    }
    for (auto df : results)
    {
        auto        p2df = (dataframe *)(PyObject_CallFunction((PyObject *)dataframe::ptype, (char*) ""));
        p2df->set(df);
        obj->dtfs.push_back(&p2df->ob_base);
        Py_IncRef(obj->dtfs.back());
    }
    return 0;
}

PyObject *p3::eval::str(PyObject *self)
{
    auto        eva = ((eval *) self)->eva;
    if (nullptr != eva) return PyUnicode_FromString(eva->name().c_str());
    PyErr_SetString(PyExc_RuntimeError, util::string_format("Initialisation error").c_str());
    return nullptr;
}

PyObject *p3::eval::evals(PyObject *, PyObject *)
{
    auto res = PyDict_New();
    auto        evals = ::Eval::evals();
    for (auto eval : evals)
    {
        auto        id = ::Eval::type2id(eval.first);
        auto        name = PyUnicode_FromString(eval.second->name().c_str());
        PyDict_SetItemString(res, id.c_str(), name);
    }
    return res;
}

PyObject *p3::eval::eval_params(PyObject *, PyObject *args)
{
    const char *type_id;
    if (! PyArg_ParseTuple(args, "s", &type_id))
        return nullptr;

    auto        eval = ::Eval::create(::Eval::type_from_id(type_id));
    if (nullptr == eval)
        return nullptr;

    auto res = PyDict_New();

    // Name
    auto        name = PyUnicode_FromString(eval->name().c_str());
    PyDict_SetItemString(res, "name", name);
    Py_DECREF(name);

    // Params
    auto        params = eval->params();
    if (params.size() > 0)
    {
        auto        lp = PyList_New(0);
        for (auto p : params)
        {
            PyObject        *parent = lp;
            auto        md = PyDict_New();
            auto        id = PyUnicode_FromString(p.first.c_str());
            PyDict_SetItemString(md, "id", id);
            Py_DECREF(id);
            auto        desc = PyUnicode_FromString(p.second.label.c_str());
            PyDict_SetItemString(md, "desc", desc);
            Py_DECREF(desc);
            auto        tp = p.second.type;
            std::string     ts = "", def = "";
            switch (tp)
            {
            case Eval::INT:
                ts = "Integer";
                def = std::to_string(p.second.default_value.value.int_val);
                break;
            case Eval::FLOAT:
                ts = "Floating-point";
                def = std::to_string(p.second.default_value.value.float_val);
                break;
            case Eval::PATH:
                ts = "File path";
                def = p.second.default_value.value.path;
                break;
            }
            auto        type = PyUnicode_FromString(ts.c_str());
            PyDict_SetItemString(md, "type", type);
            Py_DECREF(type);
            auto        dv = PyUnicode_FromString(def.c_str());
            PyDict_SetItemString(md, "default", dv);
            Py_DECREF(dv);
            PyList_Append(parent, md);
            Py_DECREF(md);
        }
        PyDict_SetItemString(res, "params", lp);
        Py_DECREF(lp);
    }
    // Metrics
    auto        metrics = eval->metrics();
    if (metrics.size() > 0)
    {
        auto        lm = PyList_New(0);
        std::map<short, PyObject*>      groups;
        for (auto m : metrics)
        {
            auto        group = m.second.group;
            PyObject        *parent = lm;
            if (-1 != group)
            {                auto        g = groups.find(group);
                if (groups.end() == g)
                {
                    parent = PyList_New(0);
                    groups[group] = parent;
                }
                else parent = g->second;
            }
            auto        md = PyDict_New();
            auto        id = PyUnicode_FromString(m.first.c_str());
            PyDict_SetItemString(md, "id", id);
            Py_DECREF(id);
            auto        desc = PyUnicode_FromString(m.second.label.c_str());
            PyDict_SetItemString(md, "desc", desc);
            Py_DECREF(desc);
            PyList_Append(parent, md);
            Py_DECREF(md);
        }
        for (auto g : groups)
        {
            PyList_Append(lm, g.second);
            Py_DECREF(g.second);
        }
        PyDict_SetItemString(res, "metrics", lm);
        Py_DECREF(lm);
    }
    return res;
}

PyObject *p3::eval::get_dtfs(PyObject *self, PyObject *)
{
    auto        ret = PyList_New(0);
    for (auto dtf : ((eval *)self)->dtfs)
    {
        PyList_Append(ret, dtf);
        Py_IncRef(dtf);
    }
    return ret;
}

PyObject *p3::eval::get_res_df(PyObject *self, PyObject *)
{
    auto df = ((eval *)self)->dtfs.front();
    Py_IncRef(df);
    return df;
}

PyObject *p3::eval::get_err(PyObject *self, PyObject *)
{
    auto        obj = (eval *) self;
    return PyUnicode_FromString(obj->err.c_str());
}
